#include <cassert>
#include <QtDebug>
#include "marchingsquare.h"

MarchingSquare::MarchingSquare(int *pmesh, int pnx, int pny, int pstep) :
        mesh(pmesh), nx(pnx), ny(pny), step(pstep)
{
    mark = new int[nx*ny];
}

MarchingSquare::~MarchingSquare() {
    delete[] mark;
}

void MarchingSquare::reverse(QList<QPoint>& list) {
    int n = list.size();
    for (int i = 0; i < n/2; i++) {
        list.swap(i, n-1-i);
    }
}

/**
 * Interpolate niveau point on line (ix0,iy0)-(ix1,iy1)
 */
QPoint MarchingSquare::msVertex(int niveau, int ix0, int iy0, int ix1, int iy1)
{
  int v0, v1;
  struct QPoint p;

  v0 = mesh[iy0*nx+ix0];
  v1 = mesh[iy1*nx+ix1];
  // (x-x0) / (x1-x0) = (v-v0) / (v1-v0)
  if (ix0 != ix1) {
    p.setX(ix0*step+(niveau-v0)*(ix1-ix0)*step/(v1-v0));
    p.setY(iy0*step);
  }
  else {
    p.setX(ix0*step);
    p.setY(iy0*step+(niveau-v0)*(iy1-iy0)*step/(v1-v0));
  }
  return p;
}

int MarchingSquare::msId(int niveau, int ix, int iy)
{
  int id;
  id = 0;
  if (mesh[(iy-1)*nx+ix-1] > niveau) id += 1;
  if (mesh[(iy-1)*nx+ix] > niveau) id += 2;
  if (mesh[iy*nx+ix] > niveau) id += 4;
  if (mesh[iy*nx+ix-1] > niveau) id += 8;
  return id;
}

/**
 * Search start square in case 5
 */
void MarchingSquare::msStart(int niveau, int &ix0, int &iy0, int &prevDir)
{
    int ix, iy;
    int id;
    int preDir;

    ix = ix0;
    iy = iy0+1;
    preDir = 2;
    while (ix > 0 && ix < nx && iy > 0 && iy < ny
            && (ix != ix0 || iy != iy0)) {
        id = msId(niveau, ix, iy);
        qDebug()<<"* search Start ("<<ix<<", "<<iy<<") "<<id<<" mark: "<<mark[iy*nx+ix];
        switch (id) {
        case 1:
        case 14:
            if (preDir == 2) {
                ix--;
                preDir = 3;
            }
            else {
                iy--;
                preDir = 0;
            }
            break;
        case 2:
        case 13:
            if (preDir == 2) {
                ix++;
                preDir = 1;
            }
            else {
                iy--;
                preDir = 0;
            }
            break;
        case 3:
        case 12:
            if (preDir == 1) {
                ix++;
            }
            else {
                ix--;
            }
            break;
        case 4:
        case 11:
            if (preDir == 3) {
                iy++;
                preDir = 2;
            }
            else {
                ix++;
                preDir = 1;
            }
            break;
        case 5:
        case 10:
            if (preDir == 0) {
                ix++;
                preDir = 1;
            }
            else if (preDir == 1) {
                iy--;
                preDir = 0;
            }
            else if (preDir == 2) {
                ix--;
                preDir = 3;
            }
            else {
                assert(preDir == 3);
                iy++;
                preDir = 2;
            }
            break;
        case 6:
        case 9:
            if (preDir == 0) {
                iy--;
            }
            else {
                iy++;
            }
            break;
        case 7:
        case 8:
            if (preDir == 1) {
                iy++;
                preDir = 2;
            }
            else {
                ix--;
                preDir = 3;
            }
            break;
        }
    }
    if (ix == ix0 && iy == iy0) {
        prevDir = 0;
        return;
    }
    ix0 = ix;
    iy0 = iy;
    if (ix <= 0) { ix0 = 1; preDir += 2; }
    if (ix >= nx) { ix0 = nx-1; preDir += 2; }
    if (iy <= 0) { iy0 = 1; preDir += 2; }
    if (iy >= ny) { iy0 = ny-1; preDir += 2; }
    prevDir = preDir % 4;
}

int MarchingSquare::msOrient(int id, int prevDir) {
    int f = 1;
    if (id >= 8) {
        f = -1;
        id = 15-id;
    }
    switch (id) {
    case 0:
        return 0;
    case 1:
        if (prevDir == 1) { // left border
            return -f;
        } else {
            assert (prevDir == 2);
            return f;
        }
    case 2:
        if (prevDir == 2) { // lower border
            return -f;
        } else {
            assert (prevDir == 3);
            return f;
        }
    case 3:
        if (prevDir == 1) { // left border
            return -f;
        } else {
            assert (prevDir == 3);
            return f;
        }
    case 4:
    case 5:
        //qDebug()<<"msOrient f "<<f<<" id "<<id<<" prevDir "<<prevDir;
        if (prevDir == 0) { // closed or upper border
            return f;
        } else if (prevDir == 1) { // left border
            return -f;
        } else if (prevDir == 2) { // lower border
            return f;
        } else { // right border
            assert(prevDir == 3);
            return -f;
        }
    case 6:
        if (prevDir == 2) { // lower border
            return -f;
        } else {
            assert(prevDir == 0);
            return f;
        }
    case 7:
        if (prevDir == 1) { // left border
            return -f;
        } else { // upper border
            assert(prevDir == 0);
            return f;
        }
    }
    return 0;
}

int MarchingSquare::getStartDir(int ix0, int iy0, int id) {
    if (id > 7) {
        id = 15-id;
    }
    if (ix0 == 1 && ix0 == 1) {
        switch (id) {
        case 1:
        case 2:
        case 6:
            return 2;
        case 3:
        case 7:
            return 1;
        default:
            assert(false);
            break;
        }
    }
    else if (ix0 == 1) {
        return 1;
    }
    else if (iy0 == 1) {
        return 2;
    }
    else {
        assert(ix0 == nx-1 || iy0 == ny-1);
        // shouldn't possible
        assert(false);
        return -1;
    }
}

const QList<QPoint> MarchingSquare::msStep(int niveau, int ix0, int iy0)
{
    QList<QPoint> list;
    QPoint pv;
    int prevDir = -1;
    int startDir = -1; // for circles
    int id, ix, iy;
    int orient = 0;

    id = msId(niveau, ix0, iy0);
    if (id == 0 || id == 15) { // nothing to do
        mark[iy0*nx+ix0] = 2;
        return list;
    }
    qDebug()<<"Step ("<<ix0<<", "<<iy0<<") "<<id<<" mark: "<<mark[iy0*nx+ix0]<<"niveau"<<niveau;
    if (id == 4 || id == 11 || id == 5 || id == 10) {
        msStart(niveau, ix0, iy0, startDir);
        id = msId(niveau, ix0, iy0);
    }
    else {
        startDir = getStartDir(ix0, iy0, id);
    }
    if (id != 4 && id != 11 && id != 5 && id != 10) {
        qDebug()<<"Borderstep ("<<ix0<<", "<<iy0<<") "<<id<<" mark: "<<mark[iy0*nx+ix0]<<"niveau"<<niveau;
    }
    switch (startDir) {
    case 0:
        pv = msVertex(niveau, ix0-1, iy0, ix0, iy0);
        break;
    case 1:
        pv = msVertex(niveau, ix0-1, iy0-1, ix0-1, iy0);
        break;
    case 2:
        pv = msVertex(niveau, ix0-1, iy0-1, ix0, iy0-1);
        break;
    case 3:
        pv = msVertex(niveau, ix0, iy0-1, ix0, iy0);
        break;
    }
    list.append(pv);
    orient = msOrient(id, startDir);
    qDebug()<<"startDir "<<startDir<<" orient "<<orient;
    prevDir = startDir;
    ix = ix0;
    iy = iy0;
    while (ix >= 1 && iy >= 1 && ix < nx && iy < ny) {
        id = msId(niveau, ix, iy);
        qDebug()<<"- Handle ("<<ix<<", "<<iy<<") "<<id<<" mark: "<<mark[iy*nx+ix]<<"niveau"<<niveau;
        assert(mark[iy*nx+ix] < 2);
        switch (id) {
        case 1:
        case 14: // lower left
            mark[iy*nx+ix] = 2;
            if (prevDir == 2) {
                prevDir = 3;
            }
            else { // prevDir == 1
                assert(prevDir == 1);
                prevDir = 0;
            }
            break;
        case 2:
        case 13: // lower right
            mark[iy*nx+ix] = 2;
            if (prevDir == 2) {
                prevDir = 1;
            }
            else {
                assert(prevDir == 3);
                prevDir = 0;
            }
            break;
        case 3:
        case 12: // horizontal edge
            mark[iy*nx+ix] = 2;
            assert(prevDir == 1 || prevDir == 3);
            // no dir change
            break;
        case 4:
        case 11: // upper right
            mark[iy*nx+ix] = 2;
            if (prevDir == 0) {
                prevDir = 1;
            }
            else {
                assert(prevDir == 3);
                prevDir = 2;
            }
            break;
        case 5:
        case 10: // double diag
            mark[iy*nx+ix] += 1;
            if (prevDir == 0) {
                prevDir = 1;
            }
            else if (prevDir == 1) {
                prevDir = 0;
            }
            else if (prevDir == 2) {
                prevDir = 3;
            }
            else {
                prevDir = 2;
            }
            break;
        case 6:
        case 9: // vertical edge
            mark[iy*nx+ix] = 2;
            // no dir change
            assert(prevDir == 0 || prevDir == 2);
            break;
        case 7:
        case 8: // upper left
            mark[iy*nx+ix] = 2;
            if (prevDir == 1) {
                prevDir = 2;
            }
            else {
                assert(prevDir == 0);
                prevDir = 3;
            }
            break;
        }
        switch (prevDir) {
        case 0:
            pv = msVertex(niveau, ix-1, iy-1, ix, iy-1);
            iy--;
            break;
        case 1:
            pv = msVertex(niveau, ix, iy-1, ix, iy);
            ix++;
            break;
        case 2:
            pv = msVertex(niveau, ix-1, iy, ix, iy);
            iy++;
            break;
        case 3:
            pv = msVertex(niveau, ix-1, iy-1, ix-1, iy);
            ix--;
            break;
        }
        list.append(pv);
        if (ix == ix0 && iy == iy0 && prevDir == startDir) { // closed, see special 1 and 2
            break;
        }
    }
    //qDebug()<<"step list: "<<list;
    assert(orient != 0);
    qDebug()<<"line "<<list[0]<<","<<list[1]<<"..."<<list.last();
    if (orient < 0) {
        reverse(list);
        qDebug()<<"revs "<<list[0]<<","<<list[1]<<"..."<<list.last();
    }
    return list;
}

const QList<QList<QPoint> > MarchingSquare::niveauLines(int val) {
    QList<QPoint> line;
    QList<QList<QPoint> > lines;
    int ix, iy;

    for (ix = 0; ix < nx*ny; ix++) {
        mark[ix] = 0;
    }
    for (iy = 1; iy < ny; iy++) {
      for (ix = 1; ix < nx; ix++) {
        //qDebug()<<"niveauLines ("<<ix<<", "<<iy<<") mark: "<<mark[iy*nx+ix];
        if (mark[iy*nx+ix] == 2) continue;
        line = msStep(val, ix, iy);
        if (line.size() > 0) {
            lines.append(line);
        }
      }
    }
    return lines;
}
