#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <tcl.h>
#include "gauss.h"
#include "rbf.h"
#include "debug.h"

int potPointCmp(const void *a, const void *b) {
  const struct IPotPoint *p0 = a;
  const struct IPotPoint *p1 = b;
  if (p0->v.y < p1->v.y)
    return -1;
  if (p0->v.y > p1->v.y)
    return 1;
  if (p0->v.x < p1->v.x)
    return -1;
  if (p0->v.x > p1->v.x)
    return 1;
  return 0;
}

static double *rbfCoeffMatrix(struct IPotPoint *p, int n) {
  double *a;
  int i, k, nn;

  nn = n+3;
  a = (double *)Tcl_Alloc(nn*nn*sizeof(double));
  for (i = 0; i < n; i++) {
    for (k = 0; k < n; k++) {
      a[k*nn+i] = a[i*nn+k] = hypot(p[i].v.x-p[k].v.x, p[i].v.y-p[k].v.y);
    }
    a[n*nn+i] = a[i*nn+n] = 1.0;
    a[(n+1)*nn+i] = a[i*nn+n+1] = p[i].v.x;
    a[(n+2)*nn+i] = a[i*nn+n+2] = p[i].v.y;
  }
  for (i = n; i < nn; i++) {
    for (k = n; k < nn; k++) {
      a[i*nn+k] = 0.0;
    }
  }
  return a;
}

static double *rbfRhs(struct IPotPoint *p, int n) {
  double *rhs;
  int i;

  rhs = (double *)Tcl_Alloc((n+3)*sizeof(double));
  for (i = 0; i < n; i++) {
    rhs[i] = p[i].pot;
  }
  for (i = 0; i < 3; i++) {
    rhs[n+i] = 0;
  }
  return rhs;
}

static bool hasPotPointDoubles(struct IPotPoint *p, int n) {
  int i;

  for (i = 1; i < n; i++) {
    if (p[i-1].v.x == p[i].v.x && p[i-1].v.y == p[i].v.y) {
      return TRUE;
    }
  }
  return FALSE;
}

static double phi(struct Ipoint v, struct IPotPoint *points,
    double *coeff, int n) {
  double sum;
  int i;

  sum = 0.0;
  for (i = 0; i < n; i++) {
    sum += coeff[i]*hypot(v.x-points[i].v.x, v.y-points[i].v.y);
  }
  sum += coeff[n] + coeff[n+1]*v.x + coeff[n+2]*v.y;
  return sum;
}

static void computePot(int *mesh, int nx, int ny,
   struct Ipoint v0, int step, struct IPotPoint *points, double *coeff, int n)
{
  int ix, iy;
  struct Ipoint v;

  for (ix = 0; ix < nx; ix++) {
    for (iy = 0; iy < ny; iy++) {
      v.x = v0.x+step*ix;
      v.y = v0.y+step*iy;
      mesh[iy*nx + ix] = (int)phi(v, points, coeff, n);
      DBG(1, ("Potential: v(%d,%d)=%d\n", v.x, v.y, mesh[iy*nx+ix]));
    }
  }
}

static void checkRbfSolution(struct IPotPoint *points, double *coeff,
  int n0, int n1)
{
  int i;
  struct Ipoint v;
  double sum;
  double pot;

  DBG(1, ("check rbf solution:\n"));
  for (i = 0; i < n0; i++) {
    v.x = points[i].v.x;
    v.y = points[i].v.y;
    pot = phi(v, points, coeff, n0+n1);
    DBG(1, ("%2d (%8d, %8d): %8.2f = %d err: %g\n",
      i, v.x, v.y, pot, points[i].pot, pot-points[i].pot));
  }
  for (i = n0; i < n0+n1; i++) {
    v.x = points[i].v.x;
    v.y = points[i].v.y;
    pot = phi(v, points, coeff, n0+n1);
    DBG(1, ("%2d (%8d, %8d): %8.2f = %d err: %g\n",
      i, v.x, v.y, pot, points[i].pot, pot-points[i].pot));
  }
  sum = 0.0;
  for (i = 0; i < n0+n1; i++) {
    sum += coeff[i];
  }
  DBG(1, ("sum constants: %g\n", sum));
  sum = 0.0;
  for (i = 0; i < n0+n1; i++) {
    sum += coeff[i]*points[i].v.x;
  }
  DBG(1, ("sum x[i]: %g\n", sum));
  sum = 0.0;
  for (i = 0; i < n0+n1; i++) {
    sum += coeff[i]*points[i].v.y;
  }
  DBG(1, ("sum y[i]: %g\n", sum));
}

bool rbfCoeff(struct RbfData *rbfData, struct IVertex *pv0,
    struct IVertex *pv1, int isClosed) {
  struct IVertex *pv;
  struct Ibox bb;
  int i, k, n0, n1;
  struct IPotPoint *p;
  double *a;
  double *coeff;
  double *rhs;
  int *mesh;
  int step, nx, ny;

  // TODO Check for intersections and self intersections
  n0 = iVertexCount(pv0);
  n1 = iVertexCount(pv1);
  p = (struct IPotPoint *)Tcl_Alloc((n0+n1)*sizeof(struct IPotPoint));
  for (i = 0, pv = pv0; i < n0; i++, pv = pv->next) {
    p[i].v = pv->v;
    p[i].pot = 0;
  }
  for (i = 0, pv = pv1; i < n1; i++, pv = pv->next) {
    p[n0+i].v = pv->v;
    p[n0+i].pot = 1000;
  }
  for (i = 0; i < n0+n1; i++) {
    for (k = i+1; k < n0+n1; k++) {
      if (lsqr(p[i].v.x - p[k].v.x, p[i].v.y - p[k].v.y) < 1000000L) {
        Tcl_Free((char *)p);
	return FALSE;
      }
    }
  }
  bb.p0 = bb.p1 = p[0].v;
  for (i = 0; i < n0+n1; i++) {
    if (p[i].v.x < bb.p0.x) bb.p0.x = p[i].v.x;
    if (p[i].v.y < bb.p0.y) bb.p0.y = p[i].v.y;
    if (p[i].v.x > bb.p1.x) bb.p1.x = p[i].v.x;
    if (p[i].v.y > bb.p1.y) bb.p1.y = p[i].v.y;
  }
  bb.p0.x -= 1000;
  bb.p0.y -= 1000;
  bb.p1.x += 1000;
  bb.p1.y += 1000;
  qsort(p, n0+n1, sizeof(struct IPotPoint), potPointCmp);
  if (hasPotPointDoubles(p, n0+n1)) {
    Tcl_Free((char *)p);
    return FALSE;
  }
  a = rbfCoeffMatrix(p, n0+n1);
  rhs = rbfRhs(p, n0+n1);
  coeff = (double *)Tcl_Alloc((n0+n1+3)*sizeof(double));
  gauss(a, coeff, rhs, n0+n1+3);
  DBG(1, ("Coefficients rbf:\n"));
  showVector(coeff, n0+n1+3);
  checkRbfSolution(p, coeff, n0, n1);
  step = 1000;
  nx = (bb.p1.x - bb.p0.x)/step+2; // ceil + 1 additional point
  ny = (bb.p1.y - bb.p0.y)/step+2; // dito
  mesh = (int *)Tcl_Alloc(nx*ny*sizeof(double));
  computePot(mesh, nx, ny, bb.p0, step, p, coeff, n0+n1);
  rbfData->mesh = mesh;
  rbfData->p0 = bb.p0;
  rbfData->nx = nx;
  rbfData->ny = ny;
  rbfData->step = 1000;
  Tcl_Free((char *)coeff);
  Tcl_Free((char *)rhs);
  Tcl_Free((char *)a);
  Tcl_Free((char *)p);
  return TRUE;
}

static int marchingSquareId(int *mesh, int nx, int ny, int niveau,
  int ix, int iy)
{
  int id;
  id = 0;
  DBG(2, ("Get Id (%d,%d): %d-%d-%d-%d niveau: %d\n", ix, iy,
    mesh[(iy-1)*nx+ix-1], mesh[(iy-1)*nx+ix], mesh[iy*nx+ix], mesh[iy*nx+ix-1],
    niveau));
  if (mesh[(iy-1)*nx+ix-1] > niveau) id += 1;
  if (mesh[(iy-1)*nx+ix] > niveau) id += 2;
  if (mesh[iy*nx+ix] > niveau) id += 4;
  if (mesh[iy*nx+ix-1] > niveau) id += 8;
  DBG(2, ("-> id %d\n", id));
  return id;
}

/**
 * Interpolate niveau point on line (ix0,iy0)-(ix1,iy1)
 */
struct IVertex* marchingSquareVertex(int *mesh, int nx, int ny, int niveau,
  struct Ipoint p0, int step, int ix0, int iy0, int ix1, int iy1)
{
  int v0, v1;
  struct Ipoint p;

  v0 = mesh[iy0*nx+ix0];
  v1 = mesh[iy1*nx+ix1];
  // (x-x0) / (x1-x0) = (v-v0) / (v1-v0)
  if (ix0 != ix1) {
    p.x = p0.x+ix0*step+(niveau-v0)*(ix1-ix0)*step/(v1-v0);
    p.y = p0.y+iy0*step;
  }
  else {
    p.x = p0.x+ix0*step;
    p.y = p0.y+iy0*step+(niveau-v0)*(iy1-iy0)*step/(v1-v0);
  }
  DBG(1, ("MarchingSquare new vertex (%d, %d) p0: (%d, %d) ix=%d iy=%d\n",
    p.x, p.y, p0.x, p0.y, ix0, iy0));
  return iNewVertex(p);
}

/**
 * Search start square in case 5
 */
static void marchingSquareStart(int *mesh, int nx, int ny, int niveau,
  int *ix0, int *iy0)
{
  int ix, iy;
  int id;
  int preDir;

  ix = *ix0;
  iy = *iy0+1;
  preDir = 2;
  while (ix > 0 && ix < nx && iy > 0 && iy < ny 
      && (ix != *ix0 || iy != *iy0)) {
    id = marchingSquareId(mesh, nx, ny, niveau, ix, iy);
    switch (id) {
    case 1:
    case 14:
      if (preDir == 2) {
        ix--;
	preDir = 3;
      }
      else {
        iy--;
	preDir = 0;
      }
      break;
    case 2:
    case 13:
      if (preDir == 2) {
        ix++;
	preDir = 1;
      }
      else {
        iy--;
	preDir = 0;
      }
      break;
    case 3:
    case 12:
      if (preDir == 1) {
        ix++;
      }
      else {
        ix--;
      }
      break;
    case 4:
    case 11:
      if (preDir == 3) {
        iy++;
	preDir = 2;
      }
      else {
        ix++;
	preDir = 1;
      }
      break;
    case 5:
    case 10:
      if (preDir == 0) {
        ix++;
	preDir = 1;
      }
      else if (preDir == 1) {
        iy++;
	preDir = 2;
      }
      else if (preDir == 2) {
        ix--;
	preDir = 3;
      }
      else {
        iy--;
	preDir = 0;
      }
      break;
    case 6:
    case 9:
      if (preDir == 0) {
        iy--;
      }
      else {
        iy++;
      }
      break;
    case 7:
    case 8:
      if (preDir == 1) {
        iy++;
	preDir = 2;
      }
      else {
        ix--;
	preDir = 3;
      }
      break;
    }
  }
  *ix0 = ix;
  *iy0 = iy;
  if (ix <= 0) *ix0 = 1;
  if (ix >= nx) *ix0 = nx-1;
  if (iy <= 0) *iy0 = 1;
  if (iy >= ny) *iy0 = ny-1;
}

static struct IVertex *marchingSquareStep(int *mesh, int *mark, int nx, int ny,
  int niveau, struct Ipoint p0, int step, int ix0, int iy0)
{
  struct IVertex *pv0 = NULL;
  struct IVertex *pv;
  int prevDir = -1;
  int id, ix, iy;

  id = marchingSquareId(mesh, nx, ny, niveau, ix0, iy0);
  if (id == 0 || id == 15) { // nothing to do
    mark[iy0*nx+ix0] = 2;
    return NULL;
  }
  if (id == 4 || id == 11 || id == 5 || id == 10) {
    marchingSquareStart(mesh, nx, ny, niveau, &ix0, &iy0);
  }
  ix = ix0;
  iy = iy0;
  prevDir = -1;
  while (ix >= 1 && iy >= 1 && ix < nx && iy < ny) {
    id = marchingSquareId(mesh, nx, ny, niveau, ix, iy);
    DBG(1, ("Handle (%d,%d) id:%d\n", ix, iy, id));
    assert(mark[iy*nx+ix] < 2);
    switch (id) {
    case 1:
    case 14: // lower left
      mark[iy*nx+ix] = 2;
      if (prevDir == -1) {
        if (ix == 1) { // left border
	  pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	    ix-1, iy-1, ix-1, iy);
	  prevDir = 1;
	}
	else { // lower border
	  pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	    ix-1, iy-1, ix, iy-1);
	  prevDir = 2;
	}
	pv0 = iAddVertex(pv0, pv);
      }
      if (prevDir == 2) {
        pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	  ix-1, iy-1, ix-1, iy);
	ix--;
	prevDir = 3;
      }
      else { // prevDir == 1
        assert(prevDir == 1);
        pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	  ix-1, iy-1, ix, iy-1);
	iy--;
	prevDir = 0;
      }
      pv0 = iAddVertex(pv0, pv);
      break;
    case 2:
    case 13: // lower right
      mark[iy*nx+ix] = 2;
      if (prevDir == -1) {
        if (iy == 1) { // lower border
	  pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	    ix-1, iy-1, ix, iy-1);
	  prevDir = 2;
	}
	else { // right border
	  pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	    ix, iy-1, ix, iy-1);
	  prevDir = 3;
	}
	pv0 = iAddVertex(pv0, pv);
      }
      if (prevDir == 2) {
        pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	  ix, iy-1, ix, iy);
	ix++;
	prevDir = 1;
      }
      else {
        assert(prevDir == 3);
        pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	  ix-1, iy-1, ix, iy-1);
	iy--;
	prevDir = 0;
      }
      pv0 = iAddVertex(pv0, pv);
      break;
    case 3:
    case 12: // horizontal edge
      mark[iy*nx+ix] = 2;
      if (prevDir == -1) {
        if (ix == 1) { // left border
	  pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	    ix-1, iy-1, ix-1, iy);
	  prevDir = 1;
	}
	else { // right border
	  pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	    ix, iy-1, ix, iy);
	  prevDir = 3;
	}
	pv0 = iAddVertex(pv0, pv);
      }
      if (prevDir == 1) {
        pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	  ix, iy-1, ix, iy);
	ix++;
      }
      else {
        assert(prevDir == 3);
        pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	  ix-1, iy-1, ix-1, iy);
	ix--;
      }
      pv0 = iAddVertex(pv0, pv);
      break;
    case 4:
    case 11: // upper right
      mark[iy*nx+ix] = 2;
      if (prevDir == -1 || prevDir == 0) { // closed curve
        pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	  ix, iy-1, ix, iy);
	ix++;
	prevDir = 1;
      }
      else {
        assert(prevDir == 3);
        pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	  ix-1, iy, ix, iy);
	iy++;
	prevDir = 2;
      }
      pv0 = iAddVertex(pv0, pv);
      break;
    case 5:
    case 10: // double diag (should not happens)
      mark[iy*nx+ix] += 1;
      if (prevDir == -1 || prevDir == 0) { // closed curve
        pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	  ix, iy-1, ix, iy);
	ix++;
	prevDir = 1;
      }
      else if (prevDir == 1) {
        pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	  ix-1, iy, ix, iy);
	iy++;
	prevDir = 2;
      }
      else if (prevDir == 2) {
        pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	  ix-1, iy-1, ix-1, iy);
	ix--;
	prevDir = 3;
      }
      else {
        pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	  ix-1, iy-1, ix, iy-1);
	iy--;
	prevDir = 0;
      }
      pv0 = iAddVertex(pv0, pv);
      break;
    case 6:
    case 9: // vertical edge
      mark[iy*nx+ix] = 2;
      if (prevDir == -1) {
        if (iy == 1) { // lower border
	  pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	    ix-1, iy-1, ix, iy-1);
	  prevDir = 2;
	}
	else { // upper border
	  pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	    ix-1, iy, ix, iy);
	  prevDir = 0;
	}
	pv0 = iAddVertex(pv0, pv);
      }
      if (prevDir == 2) {
        pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	  ix-1, iy, ix, iy);
	iy++;
      }
      else {
        assert(prevDir == 0);
        pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	  ix-1, iy-1, ix, iy-1);
	iy--;
      }
      pv0 = iAddVertex(pv0, pv);
      break;
    case 7:
    case 8: // upper left
      mark[iy*nx+ix] = 2;
      if (prevDir == -1) {
        if (ix == 1) { // left border
	  pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	    ix-1, iy-1, ix-1, iy);
	  prevDir = 1;
	}
	else { // upper border
	  pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	    ix-1, iy, ix, iy);
	  prevDir = 0;
	}
	pv0 = iAddVertex(pv0, pv);
      }
      if (prevDir == 1) {
        pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	  ix-1, iy, ix, iy);
	iy++;
	prevDir = 2;
      }
      else {
        assert(prevDir == 0);
        pv = marchingSquareVertex(mesh, nx, ny, niveau, p0, step,
	  ix-1, iy-1, ix-1, iy);
	ix--;
	prevDir = 3;
      }
      pv0 = iAddVertex(pv0, pv);
      break;
    }
    if (ix == ix0 && iy == iy0) { // closed
      break;
    }
  }
  DBG(1, ("step: pv0: %p\n", pv0));
  return pv0;
}

struct IVertex *rbfMarchingSquare(struct RbfData *rbfData, int niveau) {
  int *mesh = rbfData->mesh;
  struct Ipoint p0 = rbfData->p0;
  int nx = rbfData->nx;
  int ny = rbfData->ny;
  int step = rbfData->step;
  int i, ix, iy, id;
  int *mark;
  struct IVertex *pv = NULL;

  mark = (int *)Tcl_Alloc(nx*ny*sizeof(int));
  for (i = 0; i < nx*ny; i++) {
    mark[i] = 0;
  }
  for (iy = 1; iy < ny; iy++) {
    for (ix = 1; ix < nx; ix++) {
      if (mark[iy*nx+ix] == 2) continue;
      pv = marchingSquareStep(mesh, mark, nx, ny, niveau, p0, step, ix, iy);
      if (pv != NULL) goto finish;
    }
  }
finish:
  DBG(1, ("pv: %p\n", pv));
  Tcl_Free((char *)mark);
  return pv;
}
