#include <cmath>
#include <cstdio>
#include <QtCore>
#include <QPainter>
#include <QTimer>
#include <QtDebug>
#include <QUrl>
#include <cairo.h>
#include "marchingsquare.h"
#include "srtm.h"
#include "projection.h"

EleLine::EleLine(int elevation) :
        myElevation(elevation)
{}

Srtm::Srtm(int zoom, const QPoint& tile) :
        myZoom(zoom), myTile0(tile)
{
    myMesh = NULL;
    outDir = ".";
    srtmDir = ".";
    myMaxZoom = 15;
}

Srtm::~Srtm() {
    qDebug()<<"destructor";
}

void Srtm::requestFiles(const QStringList& fileRequests) {
    foreach (QString filename, fileRequests) {
        QString zipFilename = filename+".zip";
        filename = srtmDir+"/"+filename;
        if (QFile(filename).exists()) {
            qDebug()<<"found "<<filename;
            continue;
        }
        if (!QFile(srtmDir+"/"+zipFilename).exists()) {
            qDebug()<<"download "<<zipFilename;
            wget(QString(SRTM_URL)+zipFilename, srtmDir+"/"+zipFilename);
        }
        if (QFile(srtmDir+"/"+zipFilename).exists()) {
            unzip(zipFilename);
        }
    }
}

void Srtm::wget(QString url, QString fileName) {
    QProcess pr(this);
    pr.start("wget", QStringList()<<url<<"-O"<<fileName);
    if (!pr.waitForStarted()) {
        qDebug()<<"wget error"<<pr.readAllStandardError();
    }
    else {
        pr.closeWriteChannel();
        if (!pr.waitForFinished()) {
            qDebug()<<"wget error"<<pr.readAllStandardError();
        }
        qDebug()<<"wget"<<pr.readAllStandardError();
    }
}

void Srtm::unzip(QString fileName) {
    QProcess pr(this);
    pr.start("unzip", QStringList()<<"-d"<<srtmDir<<"-o"<<srtmDir+"/"+fileName);
    qDebug()<<"start unzip"<<fileName;
    if (!pr.waitForStarted()) {
        qDebug()<<"unzip error"<<pr.readAllStandardError();
    }
    else {
        pr.closeWriteChannel();
        if (!pr.waitForFinished()) {
            qDebug()<<"unzip error"<<pr.readAllStandardError();
        }
        qDebug()<<"unzip"<<pr.readAllStandardError();
    }
}

int* Srtm::readfile(QString filename) {
    int *mesh;
    FILE *fp;
    int ix, iy, val;

    qDebug()<<"readfile"<<filename;
    fp = fopen((srtmDir+"/"+filename).toAscii().data(), "r");
    if (fp == NULL) {
        return NULL;
    }
    mesh = new int[1201*1201];
    for (iy = 0; iy < 1201; iy++) {
        for (ix = 0; ix < 1201; ix++) {
            val = 256*(unsigned char)fgetc(fp);
            val += (unsigned char)fgetc(fp);
            if (val > 32767)
                val -= 65536;
            //qDebug()<<ix<<iy<<val;
            mesh[(1200-iy)*1201+ix] = val;
        }
    }
    fclose(fp);
    return mesh;
}

QString Srtm::fileName(const QPointF &lonLat) {
    char sx, sy;
    int x, y;
    x = (int)floor(lonLat.x());
    y = (int)floor(lonLat.y());
    sx = (x >= 0) ? 'E' : 'W';
    sy = (y >= 0) ? 'N' : 'S';
    x = abs(x);
    y = abs(y);
    return QString("%1%2%3%4.hgt").arg(sy).arg(y, 2, 10, QChar('0')).arg(sx).arg(x, 3, 10, QChar('0'));
}


int* Srtm::copy1(int *m0, int ox, int oy, int nx, int ny) {
    int *mesh = new int[nx*ny];
    int ix, iy;

    for (iy = 0; iy < ny; iy++) {
        for (ix = 0; ix < nx; ix++) {
            mesh[iy*nx+ix] = m0[(iy+oy)*2401+ix+ox];
        }
    }
    return mesh;
}

int *Srtm::copy4(int *m00, int *m10, int *m01, int *m11) {
    int *mesh = new int[2401*2401];
    int ix, iy, val;

    for (iy = 0; iy < 2401; iy++) {
        for (ix = 0; ix < 2401; ix++) {
            if (ix < 1200 && iy < 1200) val = m00[iy*1201+ix];
            if (ix < 1200 && iy >= 1200) val = m01[(iy-1200)*1201+ix];
            if (ix >= 1200 && iy < 1200) val = m10[iy*1201+ix-1200];
            if (ix >= 1200 && iy >= 1200) val = m11[(iy-1200)*1201+ix-1200];
            mesh[iy*2401+ix] = val;
        }
    }
    return mesh;
}

void Srtm::cleanData(int *mesh, int nx, int ny) {
    int ix, iy;
    int ixx, iyy;
    int val, cnt;

    for (iy = 0; iy < ny; iy++) {
        for (ix = 0; ix < nx; ix++) {
            if (mesh[iy*nx+ix] > -16384) continue;
            val = cnt = 0;
            for (iyy = iy-1; iyy <= iy+1; iyy++) {
                for (ixx = ix-1; ixx <=ix+1; ixx++) {
                    if (ixx < 0 || ixx >= nx || iyy < 0 || iyy >= ny) continue;
                    if (mesh[iyy*nx+ixx] > -16384) {
                        val += mesh[iyy*nx+ixx];
                        cnt++;
                    }
                }
            }
            mesh[iy*nx+ix] = val/cnt;
        }
    }
}

QPoint Srtm::elevationBounds(int *mesh, int nx, int ny) const {
    int ix, iy, val;
    QPoint bounds(mesh[0], mesh[0]);

    for (iy = 0; iy < ny; iy++) {
        for (ix = 0; ix < nx; ix++) {
            val = mesh[iy*nx+ix];
            if (val < bounds.x()) bounds.setX(val);
            if (val > bounds.y()) bounds.setY(val);
        }
    }
    return bounds;
}

void Srtm::createContour(const QPointF& p0, int *mAll) {
    cleanData(mAll, 2401, 2401);
    showMesh(mAll, 2401, 2401, "mall.png");
    qDebug()<<"mAll generated";
    int dx = ((int)(p0.x()*4)) % 4;
    int dy = ((int)(p0.y()*4)) % 4;
    QPointF pp0(floor(p0.x()), floor(p0.y()));
    int ox = dx*300;
    int oy = dy*300;
    myBb0 = QPointF(pp0.x()+dx*0.25, pp0.y()+dy*0.25);
    myBb1 = myBb0+QPointF(0.75, 0.75);
    qDebug()<<QString("dx: %1 dy: %2 ox: %3 oy: %4").arg(dx).arg(dy).arg(ox).arg(oy);
    myMesh = copy1(mAll, ox, oy, 901, 901);
    showMesh(myMesh, 901, 901, "mymesh.png");
    QPoint minMax = elevationBounds(myMesh, 901, 901);
    MarchingSquare ms(myMesh, 901, 901, 1000);
    int minEle = minMax.x()/10*10+10;
    int maxEle = minMax.y()/10*10-10;
    for (int niveau = minEle; niveau <= maxEle; niveau += 10) {
        // Test
        //if (niveau % 100 != 0) continue;
        qDebug()<<"Niveau"<<niveau;
        QList<QList<QPoint> > lines = ms.niveauLines(niveau);
        qDebug()<<lines.size();
        for (int i = 0; i < lines.size(); i++) {
            qDebug()<<"line "<<i<<" size: "<<lines[i].size();
            EleLine el(niveau);
            for (int k = 0; k < lines[i].size(); k++) {
                QPoint p = lines[i][k];
                QPointF lonLat(myBb0.x()+p.x()/1200000.0, myBb0.y()+p.y()/1200000.0);
                //qDebug()<<"LonLat"<<lonLat<<"Merc:"<<lonLat2Merc(lonLat);
                el.append(lonLat);
            }
            eleLines.append(el);
        }
    }
    delete [] myMesh;
}

void Srtm::drawTiles() {
    int ix, iy;
    qDebug()<<"P0: "<<myP0;
    qDebug()<<"BoundingBoxes: "<<myBb0<<", "<<myBb1;
    QPointF ll1 = lonLat2Merc(myBb1);
    QList<TileRequest> tileRequests;
    for (ix = myTile0.x(); true; ix++) {
        iy = myTile0.y();
        QPointF border = merc2Tile(QPoint(ix, iy), ll1, myZoom);
        qDebug()<<"border: "<<ix<<", "<<iy<<border;
        if (border.x() < 256)
            break;
        for (iy = myTile0.y(); true; iy--) {
            QPointF border = merc2Tile(QPoint(ix, iy), ll1, myZoom);
            qDebug()<<"border: "<<ix<<", "<<iy<<border;
            if (border.y() >= 0)
                break;
            tileRequests.append(TileRequest(myZoom, ix, iy));
        }
    }
    while (tileRequests.size() > 0) {
        drawTile(tileRequests);
    }
}

void Srtm::drawTile(QList<TileRequest> &requests) {
    int step;
    TileRequest req = requests.first();
    requests.removeFirst();
    int z = req.z();
    int ix = req.x();
    int iy = req.y();
    if (z > myMaxZoom)
        return;
    QTextStream out(stdout);
    out<<QString("Draw tile %1_%2_%3.png").arg(z).arg(ix).arg(iy)<<"\n";
    switch (z) {
    case 10:
    case 11:
        step = 100;
        break;
    case 12:
        step = 50;
        break;
    case 13:
        step = 20;
        break;
    default:
        step = 10;
        break;
    }
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, 256, 256);
    cairo_t *cr = cairo_create(surface);
    cairo_set_source_rgba(cr, 0, 0, 0, 0);
    cairo_rectangle(cr, 0, 0, 256, 256);
    cairo_fill(cr);
    cairo_select_font_face(cr, "Geogia", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_BOLD);
    cairo_set_font_size(cr, 7);
    foreach (EleLine el, eleLines) {
        if (el.elevation() % step != 0)
            continue;
        QList<QPointF> points = el.points();
        //qDebug()<<"line "<<el.elevation()<<" size: "<<points.size();
        QList<QPointF> mp;
        for (int k = 0; k < points.size(); k++) {
            QPointF lonLat = points[k];
            //qDebug()<<"LonLat"<<lonLat<<"Merc:"<<lonLat2Merc(lonLat);
            mp.append(merc2Tile(QPoint(ix, iy), lonLat2Merc(lonLat), z));
        }
        QString text("");
        double width = 0.5;
        cairo_set_source_rgba(cr, 0.5, 0.5, 0.5, 1.0);
        if (el.elevation() % 100 == 0 && z > 11) {
            width = 1.5;
            cairo_set_source_rgba(cr, 0.5, 0.5, 0.5, 0.5);
            text = QString("%1").arg(el.elevation());
        }
        else if (el.elevation() % 50 == 0 && z > 11) {
            if (z > 12) {
                cairo_set_source_rgba(cr, 0.5, 0.5, 0.5, 0.5);
                width = 1.0;
            }
            if (z > 13) {
                text = QString("%1").arg(el.elevation());
            }
        }
        drawPolyLine(cr, mp, width, text);
    }
    QString filename = QString("%1/%2_%3_%4.png").arg(outDir).arg(z).arg(ix).arg(iy);
    qDebug()<<filename;
    cairo_surface_write_to_png(surface, filename.toStdString().data());
    cairo_destroy(cr);
    cairo_surface_destroy(surface);
    for (int i = 0; i < 2; i++) {
        for (int k = 0; k < 2; k++) {
            requests.append(TileRequest(z+1, 2*ix+i, 2*iy+k));
        }
    }

}

bool Srtm::pointIsVisible(const QPointF& p) const {
    return p.x() >= 0.0 && p.x() < 256.0 && p.y() >= 0.0 && p.y() < 256.0;
}

void Srtm::drawPolyLine(cairo_t *cr, const QList<QPointF>& points, double width, QString text) {
    int i0, i1;

    i0 = 0;
    while (i0 < points.size()) {
        for (; i0 < points.size(); i0++) {
            if (pointIsVisible(points[i0]))
                break;
        }
        if (i0 == points.size())
            return;
        for (i1 = i0; i1 < points.size(); i1++) {
            if (!pointIsVisible(points[i1]))
                break;
        }
        if (i0 > 0) i0--;
        if (i1 >= points.size()) i1 = points.size()-1;
        drawLinePart(cr, points.mid(i0, i1-i0+1), width, text);
        /*
        if (i1 < points.size())
            qDebug()<<points[i0]<<points[i1];
        */
        i0 = i1+1;
    }
}

void cairoPolyline(cairo_t *cr, const QList<QPointF>& pts) {
    cairo_move_to(cr, pts[0].x(), pts[0].y());
    for (int i = 1; i < pts.size(); i++) {
        cairo_line_to(cr, pts[i].x(), pts[i].y());
    }
    cairo_stroke(cr);
}

void Srtm::drawLinePart(cairo_t *cr, const QList<QPointF>& pts, double width, QString text) {
    int i0, i1;
    //qDebug()<<"drawLinePart "<<pts.size()<<" from "<<pts[0]<<" to "<<pts[pts.size()-1]<<" "<<text;
    cairo_set_line_width(cr, width);
    if (text == QString()) {
        cairoPolyline(cr, pts);
        return;
    }
    cairo_text_extents_t te;
    cairo_text_extents(cr, text.toUtf8().data(), &te);
    double reqWidth = te.width;
    for (i0 = i1 = pts.size()/2; i0 >= 0 && i1 < pts.size(); i0--, i1++) {
        double w = hypot(pts[i1].x()-pts[i0].x(), pts[i1].y()-pts[i0].y());
        if (w > reqWidth)
            break;
    }
    if (i0 < 0 || i1 >= pts.size()) {
        cairoPolyline(cr, pts);
        return;
    }
    QPointF shift(0.5*(pts[i0].x()+pts[i1].x()), 0.5*(pts[i0].y()+pts[i1].y()));
    cairoPolyline(cr, pts.mid(0, i0+1));
    cairoPolyline(cr, pts.mid(i1, -1));
    double angle = atan2(pts[i1].y()-pts[i0].y(), pts[i1].x()-pts[i0].x());
    cairo_save(cr);
    cairo_translate(cr, shift.x(), shift.y());
    cairo_rotate(cr, angle);
    cairo_translate(cr, -te.width/2, te.height/3);
    cairo_show_text(cr, text.toUtf8().data());
    cairo_restore(cr);
}

void Srtm::showMesh(int *mesh, int nx, int ny, QString filename) {
    int ix, iy;
    int val;

    QImage pixmap(nx, ny, QImage::Format_ARGB32);
    for (iy = 0; iy <ny; iy++) {
        for (ix = 0; ix < nx; ix++) {
            val = mesh[iy*nx+ix]*256/1000;
            pixmap.setPixel(QPoint(ix, ny-1-iy), qRgb(val, val, val));
        }
    }
    pixmap.save(filename);
}

void Srtm::run() {
    if (myZoom < 10)
        return;
    qDebug()<<"run "<<myZoom;
    while (myZoom > 10) {
        myTile0 /= 2;
        myZoom--;
    }
    qDebug()<<"Request tile "<<myTile0<<" zoom "<<myZoom;
    myP0 = tile2LonLat(myTile0+QPoint(0, 1), myZoom); // lower left border;
    QPointF p0 = myP0-QPointF(0.01, 0.01); // create a small border
    QStringList fileRequests;
    fileRequests.append(fileName(p0));
    fileRequests.append(fileName(p0+QPointF(1, 0)));
    fileRequests.append(fileName(p0+QPointF(0, 1)));
    fileRequests.append(fileName(p0+QPointF(1, 1)));
    requestFiles(fileRequests);
    int *m00 = readfile(fileRequests[0]);
    int *m10 = readfile(fileRequests[1]);
    int *m01 = readfile(fileRequests[2]);
    int *m11 = readfile(fileRequests[3]);
    if (m00 != NULL && m10 != NULL && m01 != NULL && m11 != NULL) {
        showMesh(m00, 1201, 1201, fileRequests[0]+".png");
        showMesh(m10, 1201, 1201, fileRequests[1]+".png");
        showMesh(m01, 1201, 1201, fileRequests[2]+".png");
        showMesh(m11, 1201, 1201, fileRequests[3]+".png");
        int *mAll = copy4(m00, m10, m01, m11);
        delete [] m11;
        delete [] m10;
        delete [] m01;
        delete [] m00;
        createContour(p0, mAll);
        delete [] mAll;
        drawTiles();
    }
    else {
        qDebug()<<"SRTM files not found";
    }
    qDebug()<<"ready";
}

